@@ -1,6 +1,6 @@
 /*************************************************************************
- * Copyright (c) 2019-2020, NVIDIA CORPORATION. All rights reserved.
- * Modifications Copyright (c) 2019-2020 Advanced Micro Devices, Inc. All rights reserved.
+ * Copyright (c) 2019-2022, NVIDIA CORPORATION. All rights reserved.
+ * Modifications Copyright (c) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
  *
  * See LICENSE.txt for license information
  ************************************************************************/
@@ -10,17 +10,10 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <ctype.h>
-#if defined(__HIP_PLATFORM_HCC__) || defined(__HCC__) || defined(__HIPCC__)
-#include <hsa/hsa.h>
-#include <hsa/hsa_ext_amd.h>
-#endif
 #include "core.h"
 #include "nvmlwrap.h"
 #include "xml.h"
-#if defined(__HIP_PLATFORM_HCC__) || defined(__HCC__) || defined(__HIPCC__)
-#include <hsa/hsa.h>
-#include <hsa/hsa_ext_amd.h>
-#endif
+#include "rocm_smi_wrap.h"
 
 /*******************/
 /* XML File Parser */
@@ -313,12 +306,15 @@ ncclResult_t ncclTopoXmlLoadSystem(FILE* file, struct ncclXml* xml, struct ncclX
   return ncclSuccess;
 }
 
-ncclResult_t ncclTopoGetXmlFromFile(const char* xmlTopoFile, struct ncclXml* xml) {
+ncclResult_t ncclTopoGetXmlFromFile(const char* xmlTopoFile, struct ncclXml* xml, int warn) {
   FILE* file = fopen(xmlTopoFile, "r");
   if (file == NULL) {
-    WARN("Could not open XML topology file %s : %s", xmlTopoFile, strerror(errno));
+    if (warn) {
+      WARN("Could not open XML topology file %s : %s", xmlTopoFile, strerror(errno));
+    }
     return ncclSuccess;
   }
+  INFO(NCCL_GRAPH, "Loading topology file %s", xmlTopoFile);
   struct xmlHandler handlers[] = { { "system", ncclTopoXmlLoadSystem } };
   xml->maxIndex = 0;
   NCCLCHECK(xmlLoadSub(file, xml, NULL, handlers, 1));
@@ -454,8 +450,8 @@ ncclResult_t ncclTopoGetPciNode(struct ncclXml* xml, const char* busId, struct n
   NCCLCHECK(xmlFindTagKv(xml, "pci", pciNode, "busid", busId));
   if (*pciNode == NULL) {
     NCCLCHECK(xmlAddNode(xml, NULL, "pci", pciNode));
+    NCCLCHECK(xmlSetAttr(*pciNode, "busid", busId));
   }
-  NCCLCHECK(xmlSetAttr(*pciNode, "busid", busId));
   return ncclSuccess;
 }
 
@@ -476,87 +472,123 @@ ncclResult_t ncclTopoGetXmlFromSys(struct ncclXmlNode* pciNode, struct ncclXml*
   const char* busId;
   NCCLCHECK(xmlGetAttr(pciNode, "busid", &busId));
   char* path = NULL;
-  int index;
-  NCCLCHECK(xmlGetAttrIndex(pciNode, "class", &index));
-  if (index == -1) {
-    if (path == NULL) NCCLCHECK(getPciPath(busId, &path));
+  ncclDebugNoWarn = NCCL_GRAPH;
+  getPciPath(busId, &path);
+  ncclDebugNoWarn = 0;
+
+  if (path) {
     NCCLCHECK(ncclTopoSetAttrFromSys(pciNode, path, "class", "class"));
   }
+  int index;
+  ncclDebugNoWarn = NCCL_GRAPH;
+  NCCLCHECK(xmlGetAttrIndex(pciNode, "vendor", &index));
+  if (index == -1) {
+    if (path) ncclTopoSetAttrFromSys(pciNode, path, "vendor", "vendor");
+  }
+  NCCLCHECK(xmlGetAttrIndex(pciNode, "device", &index));
+  if (index == -1) {
+    if (path) ncclTopoSetAttrFromSys(pciNode, path, "device", "device");
+  }
+  NCCLCHECK(xmlGetAttrIndex(pciNode, "subsystem_vendor", &index));
+  if (index == -1) {
+    if (path) ncclTopoSetAttrFromSys(pciNode, path, "subsystem_vendor", "subsystem_vendor");
+  }
+  NCCLCHECK(xmlGetAttrIndex(pciNode, "subsystem_device", &index));
+  if (index == -1) {
+    if (path) ncclTopoSetAttrFromSys(pciNode, path, "subsystem_device", "subsystem_device");
+  }
+  ncclDebugNoWarn = 0;
   NCCLCHECK(xmlGetAttrIndex(pciNode, "link_speed", &index));
   if (index == -1) {
-    if (path == NULL) NCCLCHECK(getPciPath(busId, &path));
-    char deviceSpeedStr[MAX_STR_LEN];
-    float deviceSpeed;
-    NCCLCHECK(ncclTopoGetStrFromSys(path, "max_link_speed", deviceSpeedStr));
-    sscanf(deviceSpeedStr, "%f GT/s", &deviceSpeed);
-    char portSpeedStr[MAX_STR_LEN];
-    float portSpeed;
-    NCCLCHECK(ncclTopoGetStrFromSys(path, "../max_link_speed", portSpeedStr));
-    if (portSpeedStr[0])
-      sscanf(portSpeedStr, "%f GT/s", &portSpeed);
-    else
-      portSpeed = deviceSpeed;
-    NCCLCHECK(xmlSetAttr(pciNode, "link_speed", portSpeed < deviceSpeed ? portSpeedStr : deviceSpeedStr));
+    if (path) {
+      char deviceSpeedStr[MAX_STR_LEN];
+      float deviceSpeed;
+      NCCLCHECK(ncclTopoGetStrFromSys(path, "max_link_speed", deviceSpeedStr));
+      sscanf(deviceSpeedStr, "%f GT/s", &deviceSpeed);
+      char portSpeedStr[MAX_STR_LEN];
+      float portSpeed;
+      NCCLCHECK(ncclTopoGetStrFromSys(path, "../max_link_speed", portSpeedStr));
+      if (portSpeedStr[0])
+        sscanf(portSpeedStr, "%f GT/s", &portSpeed);
+      else
+        portSpeed = deviceSpeed;
+      NCCLCHECK(xmlSetAttr(pciNode, "link_speed", portSpeed < deviceSpeed ? portSpeedStr : deviceSpeedStr));
+    } else {
+      NCCLCHECK(xmlSetAttr(pciNode, "link_speed", ""));
+    }
   }
   NCCLCHECK(xmlGetAttrIndex(pciNode, "link_width", &index));
   if (index == -1) {
-    if (path == NULL) NCCLCHECK(getPciPath(busId, &path));
-    char strValue[MAX_STR_LEN];
-    NCCLCHECK(ncclTopoGetStrFromSys(path, "max_link_width", strValue));
-    int deviceWidth = strtol(strValue, NULL, 0);
-    NCCLCHECK(ncclTopoGetStrFromSys(path, "../max_link_width", strValue));
-    int portWidth;
-    if (strValue[0])
-      portWidth = strtol(strValue, NULL, 0);
-    else
-      portWidth = deviceWidth;
-    NCCLCHECK(xmlSetAttrInt(pciNode, "link_width", std::min(deviceWidth,portWidth)));
+    if (path) {
+      char strValue[MAX_STR_LEN];
+      NCCLCHECK(ncclTopoGetStrFromSys(path, "max_link_width", strValue));
+      int deviceWidth = strtol(strValue, NULL, 0);
+      NCCLCHECK(ncclTopoGetStrFromSys(path, "../max_link_width", strValue));
+      int portWidth;
+      if (strValue[0])
+        portWidth = strtol(strValue, NULL, 0);
+      else
+        portWidth = deviceWidth;
+      NCCLCHECK(xmlSetAttrInt(pciNode, "link_width", std::min(deviceWidth,portWidth)));
+    } else {
+      NCCLCHECK(xmlSetAttr(pciNode, "link_width", ""));
+    }
   }
   struct ncclXmlNode* parent = pciNode->parent;
   if (parent == NULL) {
-    if (path == NULL) NCCLCHECK(getPciPath(busId, &path));
+    if (path) {
+      // Save that for later in case next step is a CPU
+      char numaIdStr[MAX_STR_LEN];
+      NCCLCHECK(ncclTopoGetStrFromSys(path, "numa_node", numaIdStr));
+      // Workaround kernel bug for now
+      if (strcmp(numaIdStr, "-1") == 0) strcpy(numaIdStr, "0");
 
-    // Save that for later in case next step is a CPU
-    char numaIdStr[MAX_STR_LEN];
-    NCCLCHECK(ncclTopoGetStrFromSys(path, "numa_node", numaIdStr));
-    // Workaround kernel bug for now
-    if (strcmp(numaIdStr, "-1") == 0) strcpy(numaIdStr, "0");
-
-    // Go up one level in the PCI tree. Rewind two "/" and follow the upper PCI
-    // switch, or stop if we reach a CPU root complex.
-    int slashCount = 0;
-    int parentOffset;
-    for (parentOffset = strlen(path)-1; parentOffset>0; parentOffset--) {
-      if (path[parentOffset] == '/') {
-        slashCount++;
-        path[parentOffset] = '\0';
-        int start = parentOffset - 1;
-        while (start>0 && path[start] != '/') start--;
-        // Check whether the parent path looks like "BBBB:BB:DD.F" or not.
-        if (checkBDFFormat(path+start+1) == 0) {
-          // This a CPU root complex. Create a CPU tag and stop there.
-          struct ncclXmlNode* topNode;
-          NCCLCHECK(xmlFindTag(xml, "system", &topNode));
-          NCCLCHECK(xmlGetSubKv(topNode, "cpu", &parent, "numaid", numaIdStr));
-          if (parent == NULL) {
-            NCCLCHECK(xmlAddNode(xml, topNode, "cpu", &parent));
-            NCCLCHECK(xmlSetAttr(parent, "numaid", numaIdStr));
-          }
-        } else if (slashCount == 2) {
-          // Continue on the upper PCI switch
-          for (int i = strlen(path)-1; i>0; i--) {
-            if (path[i] == '/') {
-              NCCLCHECK(xmlFindTagKv(xml, "pci", &parent, "busid", path+i+1));
-              if (parent == NULL) {
-                NCCLCHECK(xmlAddNode(xml, NULL, "pci", &parent));
-                NCCLCHECK(xmlSetAttr(parent, "busid", path+i+1));
+      // Go up one level in the PCI tree. Rewind two "/" and follow the upper PCI
+      // switch, or stop if we reach a CPU root complex.
+      int slashCount = 0;
+      int parentOffset;
+      for (parentOffset = strlen(path)-1; parentOffset>0; parentOffset--) {
+        if (path[parentOffset] == '/') {
+          slashCount++;
+          path[parentOffset] = '\0';
+          int start = parentOffset - 1;
+          while (start>0 && path[start] != '/') start--;
+          // Check whether the parent path looks like "BBBB:BB:DD.F" or not.
+          if (checkBDFFormat(path+start+1) == 0) {
+            // This a CPU root complex. Create a CPU tag and stop there.
+            struct ncclXmlNode* topNode;
+            NCCLCHECK(xmlFindTag(xml, "system", &topNode));
+            NCCLCHECK(xmlGetSubKv(topNode, "cpu", &parent, "numaid", numaIdStr));
+            if (parent == NULL) {
+              NCCLCHECK(xmlAddNode(xml, topNode, "cpu", &parent));
+              NCCLCHECK(xmlSetAttr(parent, "numaid", numaIdStr));
+            }
+          } else if (slashCount == 2) {
+            // Continue on the upper PCI switch
+            for (int i = strlen(path)-1; i>0; i--) {
+              if (path[i] == '/') {
+                NCCLCHECK(xmlFindTagKv(xml, "pci", &parent, "busid", path+i+1));
+                if (parent == NULL) {
+                  NCCLCHECK(xmlAddNode(xml, NULL, "pci", &parent));
+                  NCCLCHECK(xmlSetAttr(parent, "busid", path+i+1));
+                }
+                break;
               }
-              break;
             }
           }
         }
+        if (parent) break;
+      }
+    } else {
+      // No information on /sys, attach GPU to unknown CPU
+      NCCLCHECK(xmlFindTagKv(xml, "cpu", &parent, "numaid", "-1"));
+      if (parent == NULL) {
+        struct ncclXmlNode* topNode;
+        NCCLCHECK(xmlFindTag(xml, "system", &topNode));
+        NCCLCHECK(xmlAddNode(xml, topNode, "cpu", &parent));
+        NCCLCHECK(xmlSetAttr(parent, "numaid", "-1"));
+        NCCLCHECK(ncclTopoGetXmlFromCpu(parent, xml));
       }
-      if (parent) break;
     }
     pciNode->parent = parent;
     parent->subs[parent->nSubs++] = pciNode;
@@ -585,7 +617,7 @@ ncclResult_t ncclTopoGetXmlFromGpu(struct ncclXmlNode* pciNode, nvmlDevice_t nvm
       NCCLCHECK(xmlGetAttr(pciNode, "busid", &busId));
       if (busId == NULL || hipDeviceGetByPCIBusId(&dev, busId) != hipSuccess) dev = -1;
     } else {
-      NCCLCHECK(wrapNvmlDeviceGetIndex(nvmlDev, (unsigned int*)&dev));
+      NCCLCHECK(ncclNvmlDeviceGetIndex(nvmlDev, (unsigned int*)&dev));
     }
     NCCLCHECK(xmlSetAttrInt(gpuNode, "dev", dev));
   }
@@ -600,7 +632,7 @@ ncclResult_t ncclTopoGetXmlFromGpu(struct ncclXmlNode* pciNode, nvmlDevice_t nvm
       CUDACHECK(hipGetDeviceProperties(&devProp, dev));
       cudaMajor = devProp.major; cudaMinor = devProp.minor;
     } else {
-      NCCLCHECK(wrapNvmlDeviceGetCudaComputeCapability(nvmlDev, &cudaMajor, &cudaMinor));
+      NCCLCHECK(ncclNvmlDeviceGetCudaComputeCapability(nvmlDev, &cudaMajor, &cudaMinor));
     }
     NCCLCHECK(xmlSetAttrInt(gpuNode, "sm", cudaMajor*10+cudaMinor));
   }
@@ -628,7 +660,7 @@ ncclResult_t ncclTopoGetXmlFromGpu(struct ncclXmlNode* pciNode, nvmlDevice_t nvm
   NCCLCHECK(xmlGetAttrInt(gpuNode, "arch", &arch.value));
 
   struct ncclXmlNode* nvlNode = NULL;
-  NCCLCHECK(xmlGetSub(pciNode, "nvlink", &nvlNode));
+  NCCLCHECK(xmlGetSub(gpuNode, "nvlink", &nvlNode));
   if (nvlNode == NULL) {
 #if defined(__HIP_PLATFORM_HCC__) || defined(__HCC__) || defined(__HIPCC__)
     const char* busId;
@@ -638,9 +670,10 @@ ncclResult_t ncclTopoGetXmlFromGpu(struct ncclXmlNode* pciNode, nvmlDevice_t nvm
     CUDACHECK(hipGetDeviceCount(&deviceCnt));
     for (int i=0; i<deviceCnt; i++) {
       if (i != dev) {
-        uint32_t link_type, hops;
-        if (hipExtGetLinkTypeAndHopCount(dev, i, &link_type, &hops) == hipSuccess) {
-          if (link_type == HSA_AMD_LINK_INFO_TYPE_XGMI && hops == 1) {
+        RSMI_IO_LINK_TYPE rsmi_type;
+        int hops, count;
+        if (rocm_smi_getLinkInfo(dev, i, &rsmi_type, &hops, &count) == ncclSuccess) {
+          if (rsmi_type == RSMI_IOLINK_TYPE_XGMI && hops == 1) {
             char busIdStr[] = "00000000:00:00.0";
             CUDACHECK(hipDeviceGetPCIBusId(busIdStr, sizeof(busIdStr), i));
             char lowerId[NVML_DEVICE_PCI_BUS_ID_BUFFER_SIZE];
@@ -652,7 +685,7 @@ ncclResult_t ncclTopoGetXmlFromGpu(struct ncclXmlNode* pciNode, nvmlDevice_t nvm
             if (nvlNode == NULL) {
               NCCLCHECK(xmlAddNode(xml, gpuNode, "xgmi", &nvlNode));
               NCCLCHECK(xmlSetAttr(nvlNode, "target", lowerId));
-              NCCLCHECK(xmlSetAttrInt(nvlNode, "count", 1));
+              NCCLCHECK(xmlSetAttrInt(nvlNode, "count", count));
             }
           }
         }
@@ -670,15 +703,15 @@ ncclResult_t ncclTopoGetXmlFromGpu(struct ncclXmlNode* pciNode, nvmlDevice_t nvm
     for (int l=0; l<maxNvLinks; ++l) {
       // Check whether we can use this NVLink for P2P
       unsigned canP2P;
-      if ((wrapNvmlDeviceGetNvLinkCapability(nvmlDev, l, NVML_NVLINK_CAP_P2P_SUPPORTED, &canP2P) != ncclSuccess) || !canP2P) continue;
+      if ((ncclNvmlDeviceGetNvLinkCapability(nvmlDev, l, NVML_NVLINK_CAP_P2P_SUPPORTED, &canP2P) != ncclSuccess) || !canP2P) continue;
 
       // Make sure the Nvlink is up. The previous call should have trained the link.
       nvmlEnableState_t isActive;
-      if ((wrapNvmlDeviceGetNvLinkState(nvmlDev, l, &isActive) != ncclSuccess) || (isActive != NVML_FEATURE_ENABLED)) continue;
+      if ((ncclNvmlDeviceGetNvLinkState(nvmlDev, l, &isActive) != ncclSuccess) || (isActive != NVML_FEATURE_ENABLED)) continue;
 
       // Try to figure out what's on the other side of the NVLink
       nvmlPciInfo_t remoteProc;
-      if (wrapNvmlDeviceGetNvLinkRemotePciInfo(nvmlDev, l, &remoteProc) != ncclSuccess) continue;
+      if (ncclNvmlDeviceGetNvLinkRemotePciInfo(nvmlDev, l, &remoteProc) != ncclSuccess) continue;
 
       // Make a lower case copy of the bus ID for calling ncclDeviceType
       // PCI system path is in lower case
@@ -715,12 +748,14 @@ ncclResult_t ncclTopoGetXmlFromGpu(struct ncclXmlNode* pciNode, nvmlDevice_t nvm
     if (index == -1) {
       const char* busId;
       NCCLCHECK(xmlGetAttr(sub, "target", &busId));
-      if (strcmp(busId, "fffffff:ffff:ff") == 0) {
+      char* path;
+      ncclDebugNoWarn = NCCL_GRAPH;
+      getPciPath(busId, &path);
+      ncclDebugNoWarn = 0;
+      if (path == NULL || strcmp(busId, "fffffff:ffff:ff") == 0) {
         // Remote NVLink device is not visible inside this VM. Assume NVSwitch.
         NCCLCHECK(xmlSetAttr(sub, "tclass", "0x068000"));
       } else {
-        char* path;
-        NCCLCHECK(getPciPath(busId, &path));
         NCCLCHECK(ncclTopoSetAttrFromSys(sub, path, "class", "tclass"));
         free(path);
       }
@@ -733,16 +768,23 @@ ncclResult_t ncclTopoGetXmlFromGpu(struct ncclXmlNode* pciNode, nvmlDevice_t nvm
 ncclResult_t ncclTopoFillGpu(struct ncclXml* xml, const char* busId, struct ncclXmlNode** gpuNode) {
   struct ncclXmlNode* node;
   NCCLCHECK(ncclTopoGetPciNode(xml, busId, &node));
+  NCCLCHECK(xmlSetAttrIfUnset(node, "class", "0x03"));
   NCCLCHECK(ncclTopoGetXmlFromSys(node, xml));
+#if defined(__HIP_PLATFORM_HCC__) || defined(__HCC__) || defined(__HIPCC__)
+  uint32_t devIndex;
+  static int rocmsmiInit = 0;
+  if (rocmsmiInit == 0) {
+    rocmsmiInit = (rocm_smi_init() != ncclSuccess) ? 2 : 1;
+  }
+  if (rocmsmiInit == 1) {
+    if (rocm_smi_getDeviceIndexByPciBusId(busId, &devIndex) != ncclSuccess) devIndex = -1;
+  }
+  NCCLCHECK(ncclTopoGetXmlFromGpu(node, NULL, xml, gpuNode));
+#else
   nvmlDevice_t nvmlDev = NULL;
-  static int nvmlInit = 0;
-  if (nvmlInit == 0) {
-    nvmlInit = (wrapNvmlSymbols() != ncclSuccess || wrapNvmlInit() != ncclSuccess) ? 2 : 1;
-  }
-  if (nvmlInit == 1) {
-    if (wrapNvmlDeviceGetHandleByPciBusId(busId, &nvmlDev) != ncclSuccess) nvmlDev = NULL;
-  }
+  if (ncclNvmlDeviceGetHandleByPciBusId(busId, &nvmlDev) != ncclSuccess) nvmlDev = NULL;
   NCCLCHECK(ncclTopoGetXmlFromGpu(node, nvmlDev, xml, gpuNode));
+#endif
   return ncclSuccess;
 }
 
@@ -785,6 +827,7 @@ ncclResult_t ncclTopoFillNet(struct ncclXml* xml, const char* pciPath, const cha
     char busId[NVML_DEVICE_PCI_BUS_ID_BUFFER_SIZE];
     strcpy(busId, pciSysPath+offset+1);
     NCCLCHECK(ncclTopoGetPciNode(xml, busId, &parent));
+    NCCLCHECK(xmlSetAttrIfUnset(parent, "class", "0x02"));
     NCCLCHECK(ncclTopoGetXmlFromSys(parent, xml));
   } else {
     // Virtual NIC, no PCI device, attach to first CPU
